@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2012-2017  Free Software Foundation, Inc.
+ * Copyright (C) 2012-2019  Free Software Foundation, Inc.
  *
  * This file is part of GNU lightning.
  *
@@ -20,15 +20,9 @@
 #ifndef _lightning_h
 #define _lightning_h
 
-#if HAVE_CONFIG_H
-# include "config.h"
-#endif
-
 #include <unistd.h>
 #include <stdlib.h>
-#if HAVE_STDINT_H
-#  include <stdint.h>
-#endif
+#include <stdint.h>
 #include <string.h>
 
 #if defined(__hpux) && defined(__hppa__)
@@ -38,14 +32,6 @@
 #  include <machine/endian.h>
 #endif
 
-#ifdef STDC_HEADERS
-#  include <stddef.h>
-#else
-#  if !defined(offsetof)
-#    define offsetof(type, field) ((char *)&((type *)0)->field - (char *)0)
-#  endif
-#endif
-
 #ifndef __WORDSIZE
 #  if defined(WORDSIZE)				/* ppc darwin */
 #    define __WORDSIZE		WORDSIZE
@@ -138,7 +124,7 @@ typedef jit_int32_t		jit_fpr_t;
 #  include <lightning/jit_mips.h>
 #elif defined(__arm__)
 #  include <lightning/jit_arm.h>
-#elif defined(__ppc__) || defined(__powerpc__)
+#elif defined(__powerpc__)
 #  include <lightning/jit_ppc.h>
 #elif defined(__sparc__)
 #  include <lightning/jit_sparc.h>
@@ -152,6 +138,8 @@ typedef jit_int32_t		jit_fpr_t;
 #  include <lightning/jit_s390.h>
 #elif defined(__alpha__)
 #  include <lightning/jit_alpha.h>
+#elif defined(__riscv)
+#  include <lightning/jit_riscv.h>
 #endif
 
 #define jit_flag_node		0x0001	/* patch node not absolute */
@@ -160,6 +148,7 @@ typedef jit_int32_t		jit_fpr_t;
 #define jit_flag_use		0x0008	/* do not remove marker label */
 #define jit_flag_synth		0x0010	/* synthesized instruction */
 #define jit_flag_head		0x1000	/* label reached by normal flow */
+#define jit_flag_varargs	0x2000	/* call{r,i} to varargs function */
 
 #define JIT_R(index)		jit_r(index)
 #define JIT_V(index)		jit_v(index)
@@ -202,6 +191,8 @@ typedef enum {
 
 #define jit_ellipsis()		_jit_ellipsis(_jit)
     jit_code_ellipsis,
+#define jit_va_push(u)		_jit_va_push(_jit,u)
+    jit_code_va_push,
 #define jit_allocai(u)		_jit_allocai(_jit,u)
 #define jit_allocar(u, v)	_jit_allocar(_jit,u,v)
     jit_code_allocai,		jit_code_allocar,
@@ -950,6 +941,7 @@ extern void _jit_putargi(jit_state_t*, jit_word_t, jit_node_t*);
 
 extern void _jit_prepare(jit_state_t*);
 extern void _jit_ellipsis(jit_state_t*);
+extern void _jit_va_push(jit_state_t*, jit_gpr_t);
 extern void _jit_pushargr(jit_state_t*, jit_gpr_t);
 extern void _jit_pushargi(jit_state_t*, jit_word_t);
 extern void _jit_finishr(jit_state_t*, jit_gpr_t);
