@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2012-2017  Free Software Foundation, Inc.
+ * Copyright (C) 2012-2019  Free Software Foundation, Inc.
  *
  * This file is part of GNU lightning.
  *
@@ -119,6 +119,18 @@ jit_init_debug(const char *progname)
 #  if defined(__alpha__)
     disasm_info.arch = bfd_arch_alpha;
     disasm_info.mach = bfd_mach_alpha_ev6;
+#  endif
+#  if defined(__hppa__)
+    disasm_info.arch = bfd_arch_hppa;
+    disasm_info.mach = bfd_mach_hppa10;
+#  endif
+#  if defined(__riscv)
+    disasm_info.arch = bfd_arch_riscv;
+#  if __WORDSIZE == 32
+    disasm_info.mach = bfd_mach_riscv32;
+#  else
+    disasm_info.mach = bfd_mach_riscv64;
+#  endif
 #  endif
     disasm_info.print_address_func = disasm_print_address;
 
@@ -213,6 +225,8 @@ jit_finish_debug(void)
 	jit_free((jit_pointer_t *)&disasm_synthetic);
     if (disasm_symbols)
 	jit_free((jit_pointer_t *)&disasm_symbols);
+    if (disasm_bfd)
+	bfd_close (disasm_bfd);
 #endif
 }
 
