@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2012-2017  Free Software Foundation, Inc.
+ * Copyright (C) 2012-2019  Free Software Foundation, Inc.
  *
  * This file is part of GNU lightning.
  *
@@ -88,7 +88,7 @@ typedef union {
 #  define _F30_REGNO			30
 #  if __WORDSIZE == 32
 #    if NEW_ABI
-#      define stack_framesize		96
+#      define stack_framesize		144
 #    else
 #      define stack_framesize		112
 #    endif
@@ -379,7 +379,11 @@ static void _nop(jit_state_t*,jit_int32_t);
 #  define BGTZ(rs,im)			hrri(MIPS_BGTZ,rs,_ZERO_REGNO,im)
 #  define BNE(rs,rt,im)			hrri(MIPS_BNE,rs,rt,im)
 #  define JALR(r0)			hrrrit(MIPS_SPECIAL,r0,0,_RA_REGNO,0,MIPS_JALR)
-#  define JR(r0)			hrrrit(MIPS_SPECIAL,r0,0,0,0,MIPS_JR)
+#  if 1 /* supports MIPS32 R6 */
+#   define JR(r0)			hrrrit(MIPS_SPECIAL,r0,0,0,0,MIPS_JALR)
+#  else /* does not support MIPS32 R6 */
+#   define JR(r0)			hrrrit(MIPS_SPECIAL,r0,0,0,0,MIPS_JR)
+#  endif
 #  define J(i0)				hi(MIPS_J,i0)
 #  define MOVZ(rd,rs,rt)		hrrrit(0,rs,rt,rd,0,MIPS_MOVZ)
 #  define comr(r0,r1)			xori(r0,r1,-1)
@@ -2968,8 +2972,13 @@ _prolog(jit_state_t *_jit, jit_node_t *node)
 	index = (_jitc->function->self.size - stack_framesize) >> STACK_SHIFT;
 #endif
 	offset = stack_framesize + index * STACK_SLOT;
-	for (; jit_arg_reg_p(index); ++index, offset += STACK_SLOT)
-	    stxi(offset	+  WORD_ADJUST, _BP_REGNO, rn(_A0 - index));
+	for (; jit_arg_reg_p(index); ++index, offset += STACK_SLOT) {
+#if NEW_ABI
+	    SD(rn(_A0 - index), offset, _BP_REGNO);
+#else
+	    stxi(offset +  WORD_ADJUST, _BP_REGNO, rn(_A0 - index));
+#endif
+	}
     }
 }
 
@@ -3014,7 +3023,7 @@ _vastart(jit_state_t *_jit, jit_int32_t r0)
 #if NEW_ABI
     if (jit_arg_reg_p(_jitc->function->vagp))
 	addi(r0, _BP_REGNO, stack_framesize + _jitc->function->vagp *
-	     sizeof(jit_word_t));
+	     sizeof(jit_int64_t));
     else
 #endif
 	addi(r0, _BP_REGNO, _jitc->function->self.size);
