@@ -1,7 +1,8 @@
-
 #if __WORDSIZE == 32
-#if defined(__ppc__)
-#define JIT_INSTR_MAX 44
+#if defined(__powerpc__)
+#if __BYTE_ORDER == __BIG_ENDIAN
+#if _CALL_SYSV
+#define JIT_INSTR_MAX 124
     0,	/* data */
     0,	/* live */
     0,	/* align */
@@ -10,8 +11,9 @@
     0,	/* #name */
     0,	/* #note */
     0,	/* label */
-    44,	/* prolog */
+    124,	/* prolog */
     0,	/* ellipsis */
+    0,	/* va_push */
     0,	/* allocai */
     0,	/* allocar */
     0,	/* arg */
@@ -24,9 +26,9 @@
     0,	/* getarg_l */
     0,	/* putargr */
     0,	/* putargi */
-    0,	/* va_start */
-    0,	/* va_arg */
-    0,	/* va_arg_d */
+    36,	/* va_start */
+    52,	/* va_arg */
+    64,	/* va_arg_d */
     0,	/* va_end */
     4,	/* addr */
     12,	/* addi */
@@ -119,15 +121,15 @@
     0,	/* ldr_l */
     0,	/* ldi_l */
     8,	/* ldxr_c */
-    8,	/* ldxi_c */
+    16,	/* ldxi_c */
     4,	/* ldxr_uc */
-    4,	/* ldxi_uc */
+    12,	/* ldxi_uc */
     4,	/* ldxr_s */
-    4,	/* ldxi_s */
+    12,	/* ldxi_s */
     4,	/* ldxr_us */
-    4,	/* ldxi_us */
+    12,	/* ldxi_us */
     4,	/* ldxr_i */
-    4,	/* ldxi_i */
+    12,	/* ldxi_i */
     0,	/* ldxr_ui */
     0,	/* ldxi_ui */
     0,	/* ldxr_l */
@@ -141,11 +143,11 @@
     0,	/* str_l */
     0,	/* sti_l */
     4,	/* stxr_c */
-    4,	/* stxi_c */
+    12,	/* stxi_c */
     4,	/* stxr_s */
-    4,	/* stxi_s */
+    12,	/* stxi_s */
     4,	/* stxr_i */
-    4,	/* stxi_i */
+    12,	/* stxi_i */
     0,	/* stxr_l */
     0,	/* stxi_l */
     8,	/* bltr */
@@ -188,10 +190,10 @@
     16,	/* bxsubi */
     12,	/* bxsubr_u */
     16,	/* bxsubi_u */
-    0,	/* jmpr */
+    8,	/* jmpr */
     4,	/* jmpi */
-    8,	/* callr */
-    16,	/* calli */
+    12,	/* callr */
+    20,	/* calli */
     0,	/* prepare */
     0,	/* pushargr */
     0,	/* pushargi */
@@ -207,93 +209,93 @@
     0,	/* retval_i */
     0,	/* retval_ui */
     0,	/* retval_l */
-    44,	/* epilog */
+    124,	/* epilog */
     0,	/* arg_f */
     0,	/* getarg_f */
     0,	/* putargr_f */
     0,	/* putargi_f */
     4,	/* addr_f */
-    12,	/* addi_f */
+    16,	/* addi_f */
     4,	/* subr_f */
-    12,	/* subi_f */
-    12,	/* rsbi_f */
+    16,	/* subi_f */
+    16,	/* rsbi_f */
     4,	/* mulr_f */
-    12,	/* muli_f */
+    16,	/* muli_f */
     4,	/* divr_f */
-    12,	/* divi_f */
+    16,	/* divi_f */
     4,	/* negr_f */
     4,	/* absr_f */
     4,	/* sqrtr_f */
     12,	/* ltr_f */
-    20,	/* lti_f */
+    24,	/* lti_f */
     16,	/* ler_f */
-    24,	/* lei_f */
+    28,	/* lei_f */
     12,	/* eqr_f */
-    20,	/* eqi_f */
+    24,	/* eqi_f */
     16,	/* ger_f */
-    24,	/* gei_f */
+    28,	/* gei_f */
     12,	/* gtr_f */
-    20,	/* gti_f */
+    24,	/* gti_f */
     16,	/* ner_f */
-    24,	/* nei_f */
+    28,	/* nei_f */
     16,	/* unltr_f */
-    24,	/* unlti_f */
+    28,	/* unlti_f */
     16,	/* unler_f */
-    24,	/* unlei_f */
+    28,	/* unlei_f */
     16,	/* uneqr_f */
-    24,	/* uneqi_f */
+    28,	/* uneqi_f */
     16,	/* unger_f */
-    24,	/* ungei_f */
+    28,	/* ungei_f */
     16,	/* ungtr_f */
-    24,	/* ungti_f */
+    28,	/* ungti_f */
     16,	/* ltgtr_f */
-    24,	/* ltgti_f */
+    28,	/* ltgti_f */
     16,	/* ordr_f */
-    24,	/* ordi_f */
+    28,	/* ordi_f */
     12,	/* unordr_f */
-    20,	/* unordi_f */
+    24,	/* unordi_f */
     12,	/* truncr_f_i */
     0,	/* truncr_f_l */
     20,	/* extr_f */
     4,	/* extr_d_f */
     4,	/* movr_f */
-    8,	/* movi_f */
+    12,	/* movi_f */
     4,	/* ldr_f */
     8,	/* ldi_f */
     4,	/* ldxr_f */
-    4,	/* ldxi_f */
+    12,	/* ldxi_f */
     4,	/* str_f */
     8,	/* sti_f */
     4,	/* stxr_f */
-    4,	/* stxi_f */
+    12,	/* stxi_f */
     8,	/* bltr_f */
-    16,	/* blti_f */
+    20,	/* blti_f */
     12,	/* bler_f */
-    20,	/* blei_f */
+    24,	/* blei_f */
     8,	/* beqr_f */
-    16,	/* beqi_f */
+    20,	/* beqi_f */
     12,	/* bger_f */
-    20,	/* bgei_f */
+    24,	/* bgei_f */
     8,	/* bgtr_f */
-    16,	/* bgti_f */
+    20,	/* bgti_f */
     8,	/* bner_f */
-    16,	/* bnei_f */
+    20,	/* bnei_f */
     12,	/* bunltr_f */
-    20,	/* bunlti_f */
+    24,	/* bunlti_f */
     8,	/* bunler_f */
-    16,	/* bunlei_f */
+    20,	/* bunlei_f */
     12,	/* buneqr_f */
-    20,	/* buneqi_f */
+    24,	/* buneqi_f */
     8,	/* bunger_f */
-    16,	/* bungei_f */
+    20,	/* bungei_f */
     12,	/* bungtr_f */
-    20,	/* bungti_f */
+    24,	/* bungti_f */
     12,	/* bltgtr_f */
-    20,	/* bltgti_f */
+    24,	/* bltgti_f */
     8,	/* bordr_f */
-    16,	/* bordi_f */
+    20,	/* bordi_f */
     8,	/* bunordr_f */
-    16,	/* bunordi_f */
+    20,	/* bunordi_f */
     0,	/* pushargr_f */
     0,	/* pushargi_f */
     0,	/* retr_f */
@@ -304,87 +306,87 @@
     0,	/* putargr_d */
     0,	/* putargi_d */
     4,	/* addr_d */
-    12,	/* addi_d */
+    24,	/* addi_d */
     4,	/* subr_d */
-    12,	/* subi_d */
-    12,	/* rsbi_d */
+    24,	/* subi_d */
+    24,	/* rsbi_d */
     4,	/* mulr_d */
-    12,	/* muli_d */
+    24,	/* muli_d */
     4,	/* divr_d */
-    12,	/* divi_d */
+    24,	/* divi_d */
     4,	/* negr_d */
     4,	/* absr_d */
     4,	/* sqrtr_d */
     12,	/* ltr_d */
-    20,	/* lti_d */
+    32,	/* lti_d */
     16,	/* ler_d */
-    24,	/* lei_d */
+    36,	/* lei_d */
     12,	/* eqr_d */
-    20,	/* eqi_d */
+    32,	/* eqi_d */
     16,	/* ger_d */
-    24,	/* gei_d */
+    36,	/* gei_d */
     12,	/* gtr_d */
-    20,	/* gti_d */
+    32,	/* gti_d */
     16,	/* ner_d */
-    24,	/* nei_d */
+    36,	/* nei_d */
     16,	/* unltr_d */
-    24,	/* unlti_d */
+    36,	/* unlti_d */
     16,	/* unler_d */
-    24,	/* unlei_d */
+    36,	/* unlei_d */
     16,	/* uneqr_d */
-    24,	/* uneqi_d */
+    36,	/* uneqi_d */
     16,	/* unger_d */
-    24,	/* ungei_d */
+    36,	/* ungei_d */
     16,	/* ungtr_d */
-    24,	/* ungti_d */
+    36,	/* ungti_d */
     16,	/* ltgtr_d */
-    24,	/* ltgti_d */
+    36,	/* ltgti_d */
     16,	/* ordr_d */
-    24,	/* ordi_d */
+    36,	/* ordi_d */
     12,	/* unordr_d */
-    20,	/* unordi_d */
+    32,	/* unordi_d */
     12,	/* truncr_d_i */
     0,	/* truncr_d_l */
     20,	/* extr_d */
     4,	/* extr_f_d */
     4,	/* movr_d */
-    8,	/* movi_d */
+    24,	/* movi_d */
     4,	/* ldr_d */
     8,	/* ldi_d */
     4,	/* ldxr_d */
-    4,	/* ldxi_d */
+    12,	/* ldxi_d */
     4,	/* str_d */
     8,	/* sti_d */
     4,	/* stxr_d */
-    4,	/* stxi_d */
+    12,	/* stxi_d */
     8,	/* bltr_d */
-    16,	/* blti_d */
+    28,	/* blti_d */
     12,	/* bler_d */
-    20,	/* blei_d */
+    32,	/* blei_d */
     8,	/* beqr_d */
-    16,	/* beqi_d */
+    32,	/* beqi_d */
     12,	/* bger_d */
-    20,	/* bgei_d */
+    32,	/* bgei_d */
     8,	/* bgtr_d */
-    16,	/* bgti_d */
+    28,	/* bgti_d */
     8,	/* bner_d */
-    16,	/* bnei_d */
+    28,	/* bnei_d */
     12,	/* bunltr_d */
-    20,	/* bunlti_d */
+    32,	/* bunlti_d */
     8,	/* bunler_d */
-    16,	/* bunlei_d */
+    28,	/* bunlei_d */
     12,	/* buneqr_d */
-    20,	/* buneqi_d */
+    32,	/* buneqi_d */
     8,	/* bunger_d */
-    16,	/* bungei_d */
+    28,	/* bungei_d */
     12,	/* bungtr_d */
-    20,	/* bungti_d */
+    32,	/* bungti_d */
     12,	/* bltgtr_d */
-    20,	/* bltgti_d */
+    32,	/* bltgti_d */
     8,	/* bordr_d */
-    16,	/* bordi_d */
+    28,	/* bordi_d */
     8,	/* bunordr_d */
-    16,	/* bunordi_d */
+    28,	/* bunordi_d */
     0,	/* pushargr_d */
     0,	/* pushargi_d */
     0,	/* retr_d */
@@ -399,12 +401,15 @@
     0,	/* movi_d_ww */
     0,	/* movr_d_w */
     0,	/* movi_d_w */
-#endif /* __ppc__ */
+#endif /* _CALL_SYV */
+#endif /* __BYTE_ORDER */
+#endif /* __powerpc__ */
 #endif /* __WORDSIZE */
 
 #if __WORDSIZE == 32
 #if defined(__powerpc__)
 #if __BYTE_ORDER == __BIG_ENDIAN
+#if !_CALL_SYSV
 #define JIT_INSTR_MAX 136
     0,	/* data */
     0,	/* live */
@@ -416,6 +421,7 @@
     0,	/* label */
     136,	/* prolog */
     0,	/* ellipsis */
+    0,	/* va_push */
     0,	/* allocai */
     0,	/* allocar */
     0,	/* arg */
@@ -803,6 +809,7 @@
     0,	/* movi_d_ww */
     0,	/* movr_d_w */
     0,	/* movi_d_w */
+#endif /* _CALL_AIX */
 #endif /* __BYTEORDER */
 #endif /* __powerpc__ */
 #endif /* __WORDSIZE */
@@ -821,6 +828,7 @@
     0,	/* label */
     148,	/* prolog */
     0,	/* ellipsis */
+    0,	/* va_push */
     0,	/* allocai */
     0,	/* allocar */
     0,	/* arg */
@@ -1226,6 +1234,7 @@
     0,	/* label */
     124,	/* prolog */
     0,	/* ellipsis */
+    0,	/* va_push */
     0,	/* allocai */
     0,	/* allocar */
     0,	/* arg */
