@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2012-2017  Free Software Foundation, Inc.
+ * Copyright (C) 2012-2019  Free Software Foundation, Inc.
  *
  * This file is part of GNU lightning.
  *
@@ -185,11 +185,11 @@ jit_register_t		_rvs[] = {
     { rc(arg) | rc(gpr) | rc(rg8) | 0,	"%rax" },
     { rc(gpr) | rc(rg8) | 10,		"%r10" },
     { rc(gpr) | rc(rg8) | 11,		"%r11" },
-    { rc(gpr) | rc(rg8) | 12,		"%r12" },
     { rc(sav) | rc(rg8) | rc(gpr) | 3,	"%rbx" },
     { rc(sav) | rc(rg8) | rc(gpr) | 13,	"%r13" },
     { rc(sav) | rc(rg8) | rc(gpr) | 14,	"%r14" },
     { rc(sav) | rc(rg8) | rc(gpr) | 15,	"%r15" },
+    { rc(sav) | rc(gpr) | rc(rg8) | 12,	"%r12" },
     { rc(arg) | rc(rg8) | rc(gpr) | 9,	"%r9" },
     { rc(arg) | rc(rg8) | rc(gpr) | 8,	"%r8" },
     { rc(arg) | rc(rg8) | rc(gpr) | 1,	"%rcx" },
@@ -507,8 +507,7 @@ _jit_retr(jit_state_t *_jit, jit_int32_t u)
     if (JIT_RET != u)
 	jit_movr(JIT_RET, u);
     /* explicitly tell it is live */
-    else
-	jit_live(JIT_RET);
+    jit_live(JIT_RET);
     jit_ret();
     jit_dec_synth();
 }
@@ -621,6 +620,14 @@ _jit_ellipsis(jit_state_t *_jit)
     jit_dec_synth();
 }
 
+void
+_jit_va_push(jit_state_t *_jit, jit_int32_t u)
+{
+    jit_inc_synth_w(va_push, u);
+    jit_pushargr(u);
+    jit_dec_synth();
+}
+
 jit_node_t *
 _jit_arg(jit_state_t *_jit)
 {
@@ -2076,6 +2083,7 @@ _emit_code(jit_state_t *_jit)
 		vaarg_d(rn(node->u.w), rn(node->v.w), jit_x87_reg_p(node->u.w));
 		break;
 	    case jit_code_live:			case jit_code_ellipsis:
+	    case jit_code_va_push:
 	    case jit_code_allocai:		case jit_code_allocar:
 	    case jit_code_arg:
 	    case jit_code_arg_f:		case jit_code_arg_d:
